from __future__ import print_function
import tornado.ioloop
import tornado.web
import ujson,json
from concurrent.futures import ThreadPoolExecutor
from tornado.concurrent import run_on_executor
from model import model_service_graph_search_one_hop,\
    model_service_graph_search_multi_hop,\
    model_service_graph_search_multi_hop_multi_edge, \
    model_service_graph_search_one_hop_multi_edge,\
    model_service_graph_search_multi_hop_common_vertexes,\
    model_service_graph_search_match_edge,\
    model_service_graph_search_match_vertex,\
    model_service_graph_search_multi_hop_multi_edge_common_vertexes
import datetime
from data_load_config import load_config
from tornado_swagger.setup import setup_swagger
from tornado_swagger.setup import export_swagger
from tornado_swagger.model import register_swagger_model

class dataServiceConfigLoad(object):
    def __init__(self):
        print("clickhouse config load start")

    def load_config(self):
        return load_config()


class DateEncoder(json.JSONEncoder):  
    def default(self, obj):  
        if isinstance(obj, datetime.datetime):  
            return obj.strftime('%Y-%m-%d %H:%M:%S')  
        elif isinstance(obj, date):  
            return obj.strftime("%Y-%m-%d")  
        else:  
            return json.JSONEncoder.default(self, obj) 


# one pop

@register_swagger_model
class modelServiceGraphSearchOneHop(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_one_hop(data, config_params)


class graphSearchOneHopHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#multi pop
@register_swagger_model
class modelServiceGraphSearchMultiHop(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop(data, config_params)


class graphSearchMultiHopHandler(tornado.web.RequestHandler):


    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """


    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#multi pop multi edge

@register_swagger_model
class modelServiceGraphSearchMultiHopMultiEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_multi_edge(data, config_params)



class graphSearchMultiHopMultiEdgeHandler(tornado.web.RequestHandler):
    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchOneHopMultiEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_one_hop_multi_edge(data, config_params)


class graphSearchOneHopMultiEdgeHandler(tornado.web.RequestHandler):
    
    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


# multi_hop_common_vertexes

@register_swagger_model
class modelServiceGraphSearchMultiHopCommonVertexes(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_common_vertexes(data, config_params)


class graphSearchMultiHopCommonVertexesHandler(tornado.web.RequestHandler):


    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
   

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchMatchEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_match_edge(data, config_params)


class graphSearchMatchEdgeHandler(tornado.web.RequestHandler):


    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """


    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge

@register_swagger_model
class modelServiceGraphSearchMatchVertex(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_match_vertex(data, config_params)

class graphSearchMatchVertexHandler(tornado.web.RequestHandler):


    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """


    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_multi_edge_common_vertexes(data, config_params)


class graphSearchMultiHopMultiEdgeCommonVertexesHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

'''
def make_app():

    data_service = dataServiceConfigLoad()
    config_params = data_service.load_config()

    model_serviceGraphSearchOnePop = modelServiceGraphSearchOneHop()

    model_serviceGraphSearchMultiHopMultiEdge = modelServiceGraphSearchMultiHopMultiEdge()

    model_serviceGraphSearchMultiHop = modelServiceGraphSearchMultiHop()

    model_serviceGraphSearchOneHopMultiEdge = modelServiceGraphSearchOneHopMultiEdge()

    model_serviceGraphSearchMultiHopCommonVertexes = modelServiceGraphSearchMultiHopCommonVertexes()

    model_serviceGraphSearchMatchEdge = modelServiceGraphSearchMatchEdge()

    model_serviceGraphSearchMatchVertex = modelServiceGraphSearchMatchVertex()

    model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes = modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes()


    return tornado.web.Application([
        (r"/graphdb/api/v1/one_hop", graphSearchOneHopHandler, {"model_service": model_serviceGraphSearchOnePop, "config_params": config_params}),
        (r"/graphdb/api/v1/multi_hop_multi_edge", graphSearchMultiHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchMultiHopMultiEdge, "config_params": config_params}),
        (r"/graphdb/api/v1/multi_hop", graphSearchMultiHopHandler, {"model_service": model_serviceGraphSearchMultiHop, "config_params": config_params}),
        (r"/graphdb/api/v1/one_hop_multi_edge", graphSearchOneHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchOneHopMultiEdge, "config_params": config_params}),
        (r"/graphdb/api/v1/multi_hop_common_vertexes", graphSearchMultiHopCommonVertexesHandler,{"model_service": model_serviceGraphSearchMultiHopCommonVertexes, "config_params": config_params}),
        (r"/graphdb/api/v1/match_edge", graphSearchMatchEdgeHandler,{"model_service": model_serviceGraphSearchMatchEdge, "config_params": config_params}),
        (r"/graphdb/api/v1/match_vertex", graphSearchMatchVertexHandler,{"model_service": model_serviceGraphSearchMatchVertex, "config_params": config_params}),
        (r"/graphdb/api/v1/multi_hop_multi_edge_common_vertexes", graphSearchMultiHopMultiEdgeCommonVertexesHandler,{"model_service": model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes, "config_params": config_params}),
    ], debug=True)
'''

class Application(tornado.web.Application):

    data_service = dataServiceConfigLoad()
    config_params = data_service.load_config()

    model_serviceGraphSearchOnePop = modelServiceGraphSearchOneHop()

    model_serviceGraphSearchMultiHopMultiEdge = modelServiceGraphSearchMultiHopMultiEdge()

    model_serviceGraphSearchMultiHop = modelServiceGraphSearchMultiHop()

    model_serviceGraphSearchOneHopMultiEdge = modelServiceGraphSearchOneHopMultiEdge()

    model_serviceGraphSearchMultiHopCommonVertexes = modelServiceGraphSearchMultiHopCommonVertexes()

    model_serviceGraphSearchMatchEdge = modelServiceGraphSearchMatchEdge()

    model_serviceGraphSearchMatchVertex = modelServiceGraphSearchMatchVertex()

    model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes = modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes()

    _routes = [
        tornado.web.url(r"/graphdb/api/v1/one_hop", graphSearchOneHopHandler, {"model_service": model_serviceGraphSearchOnePop, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/multi_hop_multi_edge", graphSearchMultiHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchMultiHopMultiEdge, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/multi_hop", graphSearchMultiHopHandler, {"model_service": model_serviceGraphSearchMultiHop, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/one_hop_multi_edge", graphSearchOneHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchOneHopMultiEdge, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/multi_hop_common_vertexes", graphSearchMultiHopCommonVertexesHandler,{"model_service": model_serviceGraphSearchMultiHopCommonVertexes, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/match_edge", graphSearchMatchEdgeHandler,{"model_service": model_serviceGraphSearchMatchEdge, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/match_vertex", graphSearchMatchVertexHandler,{"model_service": model_serviceGraphSearchMatchVertex, "config_params": config_params}),
        tornado.web.url(r"/graphdb/api/v1/multi_hop_multi_edge_common_vertexes", graphSearchMultiHopMultiEdgeCommonVertexesHandler,{"model_service": model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes, "config_params": config_params}),
    ]

    def get_routes(self):
        #print(dir(self._routes[0]))
        #print(type(self._routes[0]))
        return self._routes

    def __init__(self):
        #setup_swagger(self._routes)
        settings = {"debug": True}
        setup_swagger(self._routes,
                      swagger_url='/doc',
                      api_base_url='/graphdb/api/v1/',
                      description='',
                      api_version='1.0.0',
                      title='Journal API',
                      contact='name@domain',
                      schemes=['https'],
                      security_definitions={
                          'ApiKeyAuth': {
                              'type': 'apiKey',
                              'in': 'header',
                              'name': 'X-API-Key'
                          }
                      })
        super(Application, self).__init__(self._routes, **settings)




if __name__ == "__main__":
    #app = make_app()

    App = Application()
    #Swagger_specification = export_swagger(App.get_routes())
    #print(Swagger_specification)
    App.get_routes()
    App.listen(10010)
    tornado.ioloop.IOLoop.current().start()
    #Swagger_specification = export_swagger(App.get_routes())



from __future__ import print_function
import tornado.ioloop
import tornado.web
import ujson,json

import sys

import os

curPath = os.path.abspath(os.path.dirname(__file__))

rootPath = os.path.split(curPath)

sys.path.append(rootPath[0])
sys.path.append(rootPath[0]+"/"+rootPath[1])

from concurrent.futures import ThreadPoolExecutor
from tornado.concurrent import run_on_executor
from model import model_service_graph_search_one_hop,\
    model_service_graph_search_multi_hop,\
    model_service_graph_search_multi_hop_multi_edge, \
    model_service_graph_search_one_hop_multi_edge,\
    model_service_graph_search_multi_hop_common_vertexes,\
    model_service_graph_search_match_edge,\
    model_service_graph_search_match_vertex,\
    model_service_graph_search_multi_hop_multi_edge_common_vertexes,\
    model_service_graph_search_find_path,\
    model_service_register_graph,\
    model_service_delete_graph,\
    model_service_show_graph,\
    model_service_insert_edge,\
    model_service_insert_vertex, \
    model_service_summary_graph,\
    model_service_description_graph,\
    model_service_path_finding,\
    model_service_create_subgraph,\
    model_service_update_subgraph_by_multi_hop_multi_edge,\
    model_service_update_subgraph_by_match_edge,\
    model_service_update_subgraph_by_find_path_multi_edge,\
    model_service_destroy_subgraph,\
    model_service_metric_indegree,\
    model_service_metric_outdegree,\
    model_service_metric_degree,\
    model_service_metric_pagerank,\
    model_service_edge_match_property,\
    model_service_vertex_match_property,\
    model_service_query_edges,\
    model_service_query_vertexes,\
    model_service_time_static_subgraph

import datetime
from data_load_config import load_config
from tornado_swagger.setup import setup_swagger
from tornado_swagger.setup import export_swagger
from tornado_swagger.model import register_swagger_model


class dataServiceConfigLoad(object):
    def __init__(self):
        print("clickhouse config load start")

    def load_config(self):
        return load_config()


class DateEncoder(json.JSONEncoder):  
    def default(self, obj):  
        if isinstance(obj, datetime.datetime):  
            return obj.strftime('%Y-%m-%d %H:%M:%S')  
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")  
        else:  
            return json.JSONEncoder.default(self, obj) 


# one pop

@register_swagger_model
class modelServiceGraphSearchOneHop(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_one_hop(data, config_params)


class graphSearchOneHopHandler(tornado.web.RequestHandler):


    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):

        """
        Description end-point
        ---
        tags:
        - graphSearchOneHop
        summary: Create user
        description: This is one hop algorithm.
        operationId: examples.api.api.createUser
        produces:
        - application/json
        parameters:
        - in: body
          name: body
          description: Created user object
          required: false
          schema:
            type: object
            properties:
              step:
                type: integer
                format: int64
              start_vertex_list:
                type:
                  - "list"
              edge_name_list:
                type: list
              graph_dir:
                type: string
              direction:
                type: string
              edge_con_list:
                type: list
              userStatus:
                type: integer
                format: int32
                description: User Status
        responses:
        "201":
          description: successful operation
        """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#multi pop
@register_swagger_model
class modelServiceGraphSearchMultiHop(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop(data, config_params)


class graphSearchMultiHopHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):

        """
        Description end-point
        ---
        tags:
        - graphSearchMultiHop
        summary: Create user
        description: This can only be done by the logged in user.
        operationId: examples.api.api.createUser
        produces:
        - application/json
        parameters:
        - in: body
          name: body
          description: Created user object
          required: false
          schema:
            type: object
            properties:
              step:
                type: integer
                format: int64
              start_vertex_list:
                type:
                  - "list"
              edge_name_list:
                type: list
              graph_dir:
                type: string
              direction:
                type: string
              edge_con_list:
                type: list
              userStatus:
                type: integer
                format: int32
                description: User Status
        responses:
        "201":
          description: successful operation
        """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#multi pop multi edge

@register_swagger_model
class modelServiceGraphSearchMultiHopMultiEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_multi_edge(data, config_params)


class graphSearchMultiHopMultiEdgeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                       Description end-point
                       ---
                       tags:
                       - Vertex/Edge-Wise Operation:Search
                       summary: Multi-Hop
                       description: This is the multi-hop service.
                       operationId: examples.api.api.Multi-Hop
                       produces:
                       - application/json
                       parameters:
                       - in: body
                         name: body
                         description: multi-hop
                         required: false
                         schema:
                           type: object
                           properties:
                             step:
                               type: integer
                               default: 1
                             start_vertex_list:
                               type: list
                               default: ['10.73.28.115','10.78.55.20']
                             edge_name_list:
                               type: list
                               default: ["tcpflow", "flow"]
                             graph_name:
                               type: string
                               default: "cyber"
                             direction:
                               type: string
                               default: "forward"
                             edge_con_list_list:
                               type: list
                               default: [["downlink_length>10000", "protocol='http'"], ["record_date='2019-04-15'"]]
                             target_field_list:
                               type: list
                               default: ["record_time"]
                             only_last_step:
                               type: boolean
                               default: true
                             plus_last_vertexes:
                               type: boolean
                               default: false
                       responses:
                       "201":
                         description: successful operation
               """



        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchOneHopMultiEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_one_hop_multi_edge(data, config_params)


class graphSearchOneHopMultiEdgeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                               Description end-point
                               ---
                               tags:
                               - Vertex/Edge-Wise Operation:Search
                               summary: One Hop
                               description: This is the one hop service.
                               operationId: examples.api.api.OneHop
                               produces:
                               - application/json
                               parameters:
                               - in: body
                                 name: body
                                 description: one hop
                                 required: false
                                 schema:
                                   type: object
                                   properties:
                                     start_vertex_list:
                                       type: list
                                       default: ['10.73.28.115', '10.78.55.20']
                                     edge_name_list:
                                       type: list
                                       default: ["tcpflow", "flow"]
                                     graph_name:
                                       type: string
                                       default: "cyber"
                                     direction:
                                       type: string
                                       default: "forward"
                                     edge_con_list_list:
                                       type: list
                                       default: [["downlink_length>10000", "protocol='http'"], ["record_date='2019-04-15'"]]
                                     target_field_list:
                                       type: list
                                       default: ["record_time"]
                               responses:
                               "201":
                                 description: successful operation
                       """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


# multi_hop_common_vertexes

@register_swagger_model
class modelServiceGraphSearchMultiHopCommonVertexes(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_common_vertexes(data, config_params)


class graphSearchMultiHopCommonVertexesHandler(tornado.web.RequestHandler):


    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
   

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchMatchEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_match_edge(data, config_params)


class graphSearchMatchEdgeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                               Description end-point
                               ---
                               tags:
                               - Vertex/Edge-Wise Operation:Search
                               summary: edge matching
                               description: This is the edge matching service.
                               operationId: examples.api.api.EdgeMatching
                               produces:
                               - application/json
                               parameters:
                               - in: body
                                 name: body
                                 description: edge
                                 required: false
                                 schema:
                                   type: object
                                   properties:
                                     edge_name:
                                       type: string
                                       default: "tcpflow"
                                     graph_name:
                                       type: string
                                       default: "cyber"
                                     edge_con_list:
                                       type: list
                                       default: ["downlink_length>100000000"]
                                     target_field_list:
                                       type: list
                                       default: ['record_time', 'downlink_length']
                               responses:
                               "201":
                                 description: successful operation
                       """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge

@register_swagger_model
class modelServiceGraphSearchMatchVertex(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_match_vertex(data, config_params)


class graphSearchMatchVertexHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                       Description end-point
                                       ---
                                       tags:
                                       - Vertex/Edge-Wise Operation:Search
                                       summary: vertices matching
                                       description: This is the vertices matching service.
                                       operationId: examples.api.api.VerticesMatching
                                       produces:
                                       - application/json
                                       parameters:
                                       - in: body
                                         name: body
                                         description: vertices
                                         required: false
                                         schema:
                                           type: object
                                           properties:
                                             vertex_name:
                                               type: string
                                               default: "ip"
                                             graph_name:
                                               type: string
                                               default: "cyber_plus"
                                             vertex_con_list:
                                               type: list
                                               default: ["speed>2","service_date=='2021-01-05'"]
                                             target_field_list:
                                               type: list
                                               default: ["service_date","speed"]
                                       responses:
                                       "201":
                                         description: successful operation
                               """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_multi_edge_common_vertexes(data, config_params)


class graphSearchMultiHopMultiEdgeCommonVertexesHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                      Description end-point
                                      ---
                                      tags:
                                      - Vertex/Edge-Wise Operation:Search
                                      summary: Multi-Hop Common Vertices
                                      description: This is the multi-hop common vertices service.
                                      operationId: examples.api.api.Multi_HopCommonVertices
                                      produces:
                                      - application/json
                                      parameters:
                                      - in: body
                                        name: body
                                        description: multi-hop common vertices
                                        required: false
                                        schema:
                                          type: object
                                          properties:
                                            step:
                                              type: integer
                                              default: 1
                                            start_vertex_list:
                                              type: list
                                              default: ['10.73.28.115', '10.78.55.20']
                                            edge_name_list:
                                              type: list
                                              default: ["tcpflow", "flow"]
                                            graph_name:
                                              type: string
                                              default: "cyber"
                                            direction:
                                              type: string
                                              default: "forward"
                                            edge_con_list_list:
                                              type: list
                                              default: [["protocol='http'"], ["record_date='2019-04-15'"]]
                                      responses:
                                      "201":
                                        description: successful operation
                              """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


#one pop multi edge
@register_swagger_model
class modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_multi_hop_multi_edge_common_vertexes(data, config_params)


class graphSearchFindPathHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                             Description end-point
                                             ---
                                             tags:
                                             - Vertex/Edge-Wise Operation:Search
                                             summary: Path Finding
                                             description: This is path finding service.
                                             operationId: examples.api.api.PathFinding
                                             produces:
                                             - application/json
                                             parameters:
                                             - in: body
                                               name: body
                                               description: path
                                               required: false
                                               schema:
                                                 type: object
                                                 properties:
                                                   step_limit:
                                                     type: integer
                                                     default: 2
                                                   start_vertex:
                                                     type: string
                                                     default: '10.73.28.115'
                                                   end_vertex:
                                                     type: string
                                                     default: "10.78.55.20"
                                                   graph_name:
                                                     type: string
                                                     default: "cyber"
                                                   target_field_list:
                                                     type: list
                                                     default: ["record_time", "downlink_length"]
                                                   edge_con_list_list:
                                                     type: list
                                                     default: [["downlink_length>10000", "protocol='http'"], ["record_date='2019-04-15'"]]
                                                   edge_name_list:
                                                     type: list
                                                     default: ["tcpflow", "flow"]
                                             responses:
                                             "201":
                                               description: successful operation
                                     """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceGraphSearchFindPath(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_graph_search_find_path(data, config_params)


class RegisterGraphHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Graph-Wise Operation
                                                     summary: Register Graph
                                                     description: This is the graph registering service.
                                                     operationId: examples.api.api.RegisterGraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: graph registering
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           graph_name:
                                                             type: string
                                                             default: "cyber"
                                                           graph_cfg:
                                                             type: string
                                                             default: '{"edges":{"tcpflow":{"db":"graph","table":"tcpflow","src":"source_ip","dst":"destination_ip","fields":["record_date","record_time","protocol","destination_port","uplink_length","downlink_length"]      },"flow":{"db":"graph","table":"flow","src":"source_ip","dst":"destination_ip","fields":["record_date","record_time","destination_port","uri"]} },"vertexes":{}}'
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceRegisterGraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_register_graph(data, config_params)


class DeleteGraphHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def delete(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Graph-Wise Operation
                                                     summary: Delete Graph
                                                     description: This is the graph deleting service.
                                                     operationId: examples.api.api.DeleteGraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: graph deleting
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           graph_name:
                                                             type: string
                                                             default: "cyber"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceDeleteGraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_delete_graph(data, config_params)


class ShowGraphHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Graph-Wise Operation
                                                     summary: Show Graph
                                                     description: This is the graph showing service.
                                                     operationId: examples.api.api.ShowGraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: graph showing
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceShowGraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_show_graph(config_params)


class InsertEdgeHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Insert
                                                     summary: Insert Edge
                                                     description: This is the edge inserting service.
                                                     operationId: examples.api.api.InsertEdge
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: insert edge
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "tcpflow"
                                                           graph_name:
                                                             type: string
                                                             default: "cyber_plus"
                                                           edge_schema:
                                                             type: list
                                                             default: ["record_time", "record_date", "source_ip", "destination_ip", "protocol", "destination_port", "uplink_length", "downlink_length"]
                                                           edge_data:
                                                             type: list
                                                             default: [["2019-04-11 18:48:59", "2019-04-11", "10.66.18.32", "184.173.90.200", "http", "80", 14725, 3116]]
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceInsertEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_insert_edge(data, config_params)


class InsertVertexHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Insert
                                                     summary: Insert Vertex
                                                     description: This is the vertex inserting service.
                                                     operationId: examples.api.api.InsertVertex
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: insert vertex
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           vertex_name:
                                                             type: string
                                                             default: "ip"
                                                           graph_name:
                                                             type: string
                                                             default: "cyber_plus"
                                                           vertex_schema:
                                                             type: list
                                                             default: ["service_date", "ip", "host", "speed"]
                                                           vertex_data:
                                                             type: list
                                                             default: [["2021-01-04","1.1.1.1","p47708v.hulk.shbt.qihoo.net","2"],["2021-01-05","1.1.1.2","p47709v.hulk.shbt.qihoo.net","3"]] 
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """


        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceInsertVertex(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_insert_vertex(data, config_params)


class SummaryGraphHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self, graph_name):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Graph-Wise Operation
                                                     summary: Summary Graph
                                                     description: This is the graph summary service.
                                                     operationId: examples.api.api.SummaryGraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: graph summary
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           graph_name:
                                                             type: string
                                                             default: "cyber"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model(graph_name)
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self, graph_name):
        #data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(graph_name, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceSummaryGraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_summary_graph(data, config_params)


class DescriptionGraphHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self, graph_name):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Graph-Wise Operation
                                                     summary: Description Graph
                                                     description: This is the graph description service.
                                                     operationId: examples.api.api.DescriptionGraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: graph description
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           graph_name:
                                                             type: string
                                                             default: "cyber"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model(graph_name)
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self, graph_name):
        # data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(graph_name, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceDescriptionGraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_description_graph(data, config_params)


class HealthServiceHandler(tornado.web.RequestHandler):
    def initialize(self):
        pass

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
               Description end-point
                ---
                tags:
                - health
                summary: health
                description: This is the health
                operationId: examples.api.api.health
                produces:
                - application/json
                parameters:
                - in: body
                  name: body
                  description: health
                  required: false
                  schema:
                    type: object
                    properties:
                      health:
                        type: str
                        default:  "health"
                responses:
                "201":
                  description: successful operation
        """
        res = yield self.data_process_and_model()
        self.write(ujson.dumps(res, ensure_ascii=False))
        self.finish("")

    @run_on_executor
    def data_process_and_model(self):
        json_datas = self.request.body.decode('utf8')
        result = {}
        result["status"] = "UP"
        return result


class InfoServiceHandler(tornado.web.RequestHandler):
    def initialize(self):
        pass

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
               Description end-point
                ---
                tags:
                - info
                summary: info
                description: This is the info
                operationId: examples.api.api.info
                produces:
                - application/json
                parameters:
                - in: body
                  name: body
                  description: info
                  required: false
                  schema:
                    type: object
                    properties:
                      info:
                        type: str
                        default:  "info"
                responses:
                "201":
                  description: successful operation
        """
        res = yield self.data_process_and_model()
        self.write(ujson.dumps(res, ensure_ascii=False))
        self.finish("")

    @run_on_executor
    def data_process_and_model(self):
        json_datas = self.request.body.decode('utf8')
        result = {}
        build = {}
        build["name"] = "graph-operation"
        build["time"] = "2021-01-29T20:20:41.578Z"
        build["version"] = "0.0.1-SNAPSHOT"
        result["build"] = build
        return result


class FindPathHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation: Path
                                                     summary: Path
                                                     description: This is the path service.
                                                     operationId: examples.api.api.Path
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: path
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           step_limit:
                                                             type: integer
                                                             default: "2"
                                                           graph_name:
                                                             type: string
                                                             default: "cyber_plus"
                                                           start_vertex_list:
                                                             type: list
                                                             default: ["1000007","100000","100"]
                                                           end_vertex_list:
                                                             type: list
                                                             default: ["343965","131254","51242","343965"]
                                                           edge_name_list:
                                                             type: list
                                                             default: ["userid_adgroup","adgroup_customer"]
                                                           edge_con_list_list:
                                                             type: list
                                                             default: [["record_date='2097-5-13'"],["record_date='2017-5-9'"]]
                                                           target_field_list:
                                                             type: list
                                                             default: [["record_date='2097-5-13'"],["record_date='2017-5-9'"]]
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """


        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServicePathFinding(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_path_finding(data, config_params)


class SubgraphCreationHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Subgraph
                                                     summary: Subgraph Creation
                                                     description: This is the subgraph creation service.
                                                     operationId: examples.api.api.SubgraphCreation
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: path
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           subgraph_name:
                                                             type: string
                                                             default: "taobao_sub"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceSubgraphCreation(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_create_subgraph(data, config_params)


class MultiHopSubgraphUpdateHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)



    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                               Description end-point
                               ---
                               tags:
                               - Subgraph
                               summary: Subgraph Update Multi-Hop
                               description: This is the multi-hop subgraph update service.
                               operationId: examples.api.api.SubgraphUpdateMultiHop
                               produces:
                               - application/json
                               parameters:
                               - in: body
                                 name: body
                                 description: subgraph update multi-hop
                                 required: false
                                 schema:
                                   type: object
                                   properties:
                                     step:
                                       type: integer
                                       default: 1
                                     start_vertex_list:
                                       type: list
                                       default: ['10.73.28.115','10.78.55.20']
                                     edge_name_list:
                                       type: list
                                       default: ["tcpflow", "flow"]
                                     graph_name:
                                       type: string
                                       default: "taobao"
                                     subgraph_name:
                                       type: string
                                       default: "taobao_sub"
                                     direction:
                                       type: string
                                       default: "forward"
                                     edge_con_list_list:
                                       type: list
                                       default: [["downlink_length>10000", "protocol='http'"], ["record_date='2019-04-15'"]]
                               responses:
                               "201":
                                 description: successful operation
                       """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceMultiHopSubgraphUpdate(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_update_subgraph_by_multi_hop_multi_edge(data, config_params)


class PathSubgraphUpdateHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                             Description end-point
                                                             ---
                                                             tags:
                                                             - Subgraph
                                                             summary: Subgraph Update Path
                                                             description: This is the path subgraph update service.
                                                             operationId: examples.api.api.SubgraphUpdatePath
                                                             produces:
                                                             - application/json
                                                             parameters:
                                                             - in: body
                                                               name: body
                                                               description: path
                                                               required: false
                                                               schema:
                                                                 type: object
                                                                 properties:
                                                                   step_limit:
                                                                     type: integer
                                                                     default: "2"
                                                                   graph_name:
                                                                     type: string
                                                                     default: "taobao"
                                                                   subgraph_name:
                                                                     type: string
                                                                     default: "taobao_sub"
                                                                   start_vertex_list:
                                                                     type: list
                                                                     default: ["1000007","100000","100"]
                                                                   end_vertex_list:
                                                                     type: list
                                                                     default: ["343965","131254","51242","343965"]
                                                                   edge_name_list:
                                                                     type: list
                                                                     default: ["userid_adgroup","adgroup_customer"]
                                                                   edge_con_list_list:
                                                                     type: list
                                                                     default: [["record_date='2097-5-13'"],["record_date='2017-5-9'"]]
                                                             responses:
                                                             "201":
                                                               description: successful operation
                                                     """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServicePathSubgraphUpdate(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_update_subgraph_by_find_path_multi_edge(data, config_params)


class EdgeSubgraphUpdateHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                       Description end-point
                                       ---
                                       tags:
                                       - Subgraph
                                       summary: Subgraph Update Edge
                                       description: This is the edge subgraph update service.
                                       operationId: examples.api.api.SubgraphUpdateEdge
                                       produces:
                                       - application/json
                                       parameters:
                                       - in: body
                                         name: body
                                         description: subgraph update edge
                                         required: false
                                         schema:
                                           type: object
                                           properties:
                                             edge_name:
                                               type: string
                                               default: "tcpflow"
                                             graph_name:
                                               type: string
                                               default: "taobao"
                                             subgraph_name:
                                               type: string
                                               default: "taobao_sub"
                                             edge_con_list:
                                               type: list
                                               default: ["downlink_length>100000000"]
                                       responses:
                                       "201":
                                         description: successful operation
                               """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceEdgeSubgraphUpdate(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_update_subgraph_by_match_edge(data, config_params)


class SubgraphDestroyHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def post(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Subgraph
                                                     summary: Subgraph Destroy
                                                     description: This is the subgraph destroy update service.
                                                     operationId: examples.api.api.SubgraphDestroy
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: Subgraph Destroy
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           subgraph_name:
                                                             type: string
                                                             default: "taobao_sub"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceSubgraphDestroy(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_destroy_subgraph(data, config_params)


class MetricIndegreeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - vertex/edge-wise operation: metric
                                                     summary: Metric Indegree
                                                     description: This is the metric indegree service.
                                                     operationId: examples.api.api.MetricIndegree
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: Metric Indegree
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "user_adgroup"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                           if_sort:
                                                             type: boolean
                                                             default: False
                                                           topk:
                                                             type: integer
                                                             default: -1
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceMetricIndegree(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_metric_indegree(data, config_params)


class MetricOutdegreeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - vertex/edge-wise operation: metric
                                                     summary: Metric Outdegree
                                                     description: This is the metric outdegree service.
                                                     operationId: examples.api.api.MetricOutdegree
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: Metric Outdegree
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "user_adgroup"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                           if_sort:
                                                             type: boolean
                                                             default: False
                                                           topk:
                                                             type: integer
                                                             default: -1
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceMetricOutdegree(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_metric_outdegree(data, config_params)


class MetricDegreeHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - vertex/edge-wise operation: metric
                                                     summary: Metric Degree
                                                     description: This is the metric degree service.
                                                     operationId: examples.api.api.MetricDegree
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: Metric Degree
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "user_adgroup"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                           if_sort:
                                                             type: boolean
                                                             default: False
                                                           topk:
                                                             type: integer
                                                             default: -1
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceMetricDegree(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_metric_degree(data, config_params)


class MetricPagerankHandler(tornado.web.RequestHandler):

    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - vertex/edge-wise operation: metric
                                                     summary: Metric Pagerank
                                                     description: This is the metric pagerank service.
                                                     operationId: examples.api.api.MetricPagerank
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: Metric Pagerank
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "user_adgroup"
                                                           graph_name:
                                                             type: string
                                                             default: "taobao"
                                                           if_sort:
                                                             type: boolean
                                                             default: False
                                                           topk:
                                                             type: integer
                                                             default: -1
                                                           d:
                                                             type: float
                                                             default: 0.85
                                                           num_iter:
                                                             type: integer
                                                             default: 10
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """
        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceMetricPagerank(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_metric_pagerank(data, config_params)


class EdgeMatchPropertyHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Match
                                                     summary: Match Edge Property
                                                     description: This is the edge Property matching service.
                                                     operationId: examples.api.api.EdgePropertyMatching
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: edge property matching
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "transactions"
                                                           graph_name:
                                                             type: string
                                                             default: "anti_money_launder"
                                                           start_vertex_list:
                                                             type: list
                                                             default: ["770305"]
                                                           end_vertex_list:
                                                             type: list
                                                             default: ["402592"]
                                                           egde_con_list:
                                                             type: list
                                                             default: ["record_date < \'2017-05-11\'"]
                                                           target_field_list:
                                                             type: list
                                                             default: []
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceEdgeMatchProperty(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_edge_match_property(data, config_params)


class VertexMatchPropertyHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Match
                                                     summary: Match Vertex Property
                                                     description: This is the vertex property matching service.
                                                     operationId: examples.api.api.VertexPropertyMatching
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: vertex property matching
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           vertex_name:
                                                             type: string
                                                             default: "account"
                                                           graph_name:
                                                             type: string
                                                             default: "anti_money_launder"
                                                           vertex_id_list:
                                                             type: list
                                                             default: ['119309','151550']
                                                           vertex_con_list:
                                                             type: list
                                                             default: ["record_date < \'2017-05-11\'"]
                                                           target_field_list:
                                                             type: list
                                                             default: None
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceVertexMatchProperty(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_vertex_match_property(data, config_params)


class EdgeQueryHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Match
                                                     summary: Match Edge Property
                                                     description: This is the edge Property matching service.
                                                     operationId: examples.api.api.EdgePropertyMatching
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: edge property matching
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "transactions"
                                                           graph_name:
                                                             type: string
                                                             default: "anti_money_launder"
                                                           edge_con_list:
                                                             type: list
                                                             default: ["record_date < \'2017-05-11\'"]
                                                           target_field_list:
                                                             type: list
                                                             default: []
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceQueryEdge(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_query_edges(data, config_params)


class VertexQueryHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Match
                                                     summary: Match Vertex Property
                                                     description: This is the vertex property matching service.
                                                     operationId: examples.api.api.VertexPropertyMatching
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: vertex property matching
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           vertex_name:
                                                             type: string
                                                             default: "account"
                                                           graph_name:
                                                             type: string
                                                             default: "anti_money_launder"
                                                           vertex_con_list:
                                                             type: list
                                                             default: ["record_date < \'2017-05-11\'"]
                                                           target_field_list:
                                                             type: list
                                                             default: None
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result

@register_swagger_model
class modelServiceQueryVertex(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_query_vertexes(data, config_params)


class TimeStaticSubgraphHandler(tornado.web.RequestHandler):

    """
        Description end-point
        ---
        tags:
        - Example
        summary: Create user
    """
    def initialize(self, model_service, config_params):
        self.model = model_service
        self.config_params = config_params

    executor = ThreadPoolExecutor(5)

    @tornado.web.asynchronous
    @tornado.gen.coroutine
    def get(self):
        """
                                                     Description end-point
                                                     ---
                                                     tags:
                                                     - Vertex/Edge-Wise Operation:Statistic
                                                     summary: Time Static Subgraph
                                                     description: This is the time static subgraph service.
                                                     operationId: examples.api.api.TimeStaticSubgraph
                                                     produces:
                                                     - application/json
                                                     parameters:
                                                     - in: body
                                                       name: body
                                                       description: time static subgraph
                                                       required: false
                                                       schema:
                                                         type: object
                                                         properties:
                                                           edge_name:
                                                             type: string
                                                             default: "account"
                                                           graph_name:
                                                             type: string
                                                             default: "anti_money_launder"
                                                           subgraph_name:
                                                             type: string
                                                             default: "sub_anti_money_laundry2"
                                                           edge_con_list:
                                                             type: list
                                                             default: ["record_date < '2017-05-11'"]
                                                           time_dimention:
                                                             type: string
                                                             default: "Minute"
                                                     responses:
                                                     "201":
                                                       description: successful operation
                                             """

        res = yield self.data_process_and_model()
        self.write(json.dumps(res, cls=DateEncoder))
        self.finish()

    @run_on_executor
    def data_process_and_model(self):
        data_json = ujson.loads(self.request.body)
        result_list = self.model.model_operation(data_json, self.config_params)
        result = {
            "result": result_list
        }
        self.set_header("Content-Type", "application/json; charset=UTF-8")
        return result


@register_swagger_model
class modelServiceTimeStaticSubgraph(object):
    def __init__(self):
        print("clickhouse model start")

    def model_operation(self, data, config_params):
        return model_service_time_static_subgraph(data, config_params)



class Application(tornado.web.Application):

    data_service = dataServiceConfigLoad()

    config_params = data_service.load_config()

    model_serviceGraphSearchOnePop = modelServiceGraphSearchOneHop()

    model_serviceGraphSearchMultiHopMultiEdge = modelServiceGraphSearchMultiHopMultiEdge()

    model_serviceGraphSearchMultiHop = modelServiceGraphSearchMultiHop()

    model_serviceGraphSearchOneHopMultiEdge = modelServiceGraphSearchOneHopMultiEdge()

    model_serviceGraphSearchMultiHopCommonVertexes = modelServiceGraphSearchMultiHopCommonVertexes()

    model_serviceGraphSearchMatchEdge = modelServiceGraphSearchMatchEdge()

    model_serviceGraphSearchMatchVertex = modelServiceGraphSearchMatchVertex()

    model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes = modelServiceGraphSearchMultiHopMultiEdgeCommonVertexes()

    model_serviceGraphSearchFindPath = modelServiceGraphSearchFindPath()

    model_serviceRegisterGraph = modelServiceRegisterGraph()

    model_serviceDeleteGraph = modelServiceDeleteGraph()

    model_serviceShowGraph = modelServiceShowGraph()

    model_serviceInsertEdge = modelServiceInsertEdge()

    model_serviceInsertVertex = modelServiceInsertVertex()

    model_serviceSummaryGraph = modelServiceSummaryGraph()

    model_serviceDescriptionGraph = modelServiceDescriptionGraph()

    model_serviceFindPath = modelServicePathFinding()

    model_serviceSubgraphCreation = modelServiceSubgraphCreation()

    model_serviceMultiHopSubgraphUpdate = modelServiceMultiHopSubgraphUpdate()

    model_servicePathSubgraphUpdate = modelServicePathSubgraphUpdate()

    model_serviceEdgeSubgraphUpdate = modelServiceEdgeSubgraphUpdate()

    model_serviceSubgraphDestroy = modelServiceSubgraphDestroy()

    model_serviceMetricIndegree = modelServiceMetricIndegree()

    model_serviceMetricOutdegree = modelServiceMetricOutdegree()

    model_serviceMetricDegree = modelServiceMetricDegree()

    model_serviceMetricPagerank = modelServiceMetricPagerank()

    model_serviceEdgeMatchProperty = modelServiceEdgeMatchProperty()

    model_serviceVertexMatchProperty = modelServiceVertexMatchProperty()

    model_serviceQueryEdge = modelServiceQueryEdge()

    model_serviceQueryVertex = modelServiceQueryVertex()

    model_serviceTimeStaticSubgraph = modelServiceTimeStaticSubgraph()

    _routes = [
        tornado.web.url(r"/graph-db/api/v1/multi-hop", graphSearchMultiHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchMultiHopMultiEdge, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/one-hop", graphSearchOneHopMultiEdgeHandler, {"model_service": model_serviceGraphSearchOneHopMultiEdge, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/edges", graphSearchMatchEdgeHandler,{"model_service": model_serviceGraphSearchMatchEdge, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/vertices", graphSearchMatchVertexHandler,{"model_service": model_serviceGraphSearchMatchVertex, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/multi-hop-common-vertices", graphSearchMultiHopMultiEdgeCommonVertexesHandler, {"model_service": model_serviceGraphSearchMultiHopMultiEdgeCommonVertexes, "config_params": config_params}),
        # tornado.web.url(r"/graph-db/api/v1/paths", graphSearchFindPathHandler,{"model_service": model_serviceGraphSearchFindPath,"config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/graph-registration", RegisterGraphHandler,{"model_service": model_serviceRegisterGraph, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/graph-deletion", DeleteGraphHandler,{"model_service": model_serviceDeleteGraph, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/graph-show", ShowGraphHandler,{"model_service": model_serviceShowGraph, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/graph-summary/(?P<graph_name>\S*)", SummaryGraphHandler,{"model_service": model_serviceSummaryGraph, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/graph-description/(?P<graph_name>\S*)", DescriptionGraphHandler,{"model_service": model_serviceDescriptionGraph, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/edge-insertion", InsertEdgeHandler,{"model_service": model_serviceInsertEdge, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/vertex-insertion", InsertVertexHandler,{"model_service": model_serviceInsertVertex, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/health", HealthServiceHandler),
        tornado.web.url(r"/graph-db/api/v1/info", InfoServiceHandler),
        tornado.web.url(r"/graph-db/api/v1/path", FindPathHandler, {"model_service": model_serviceFindPath, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-creation", SubgraphCreationHandler,{"model_service": model_serviceSubgraphCreation, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-update-multi-hop", MultiHopSubgraphUpdateHandler,{"model_service": model_serviceMultiHopSubgraphUpdate, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-update-path", PathSubgraphUpdateHandler, {"model_service": model_servicePathSubgraphUpdate, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-update-edge", EdgeSubgraphUpdateHandler, {"model_service": model_serviceEdgeSubgraphUpdate, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-destruction", SubgraphDestroyHandler, {"model_service": model_serviceSubgraphDestroy, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/metric-indegree", MetricIndegreeHandler, {"model_service": model_serviceMetricIndegree, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/metric-outdegree", MetricOutdegreeHandler, {"model_service": model_serviceMetricOutdegree, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/metric-degree", MetricDegreeHandler, {"model_service": model_serviceMetricDegree, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/metric-pagerank", MetricPagerankHandler, {"model_service": model_serviceMetricPagerank, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/vertex-property", VertexMatchPropertyHandler,{"model_service": model_serviceVertexMatchProperty, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/edge-property", EdgeMatchPropertyHandler, {"model_service": model_serviceEdgeMatchProperty, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/vertex-query", VertexQueryHandler, {"model_service": model_serviceQueryVertex,"config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/edge-query", EdgeQueryHandler, {"model_service": model_serviceQueryEdge, "config_params": config_params}),
        tornado.web.url(r"/graph-db/api/v1/subgraph-time-statistic", TimeStaticSubgraphHandler,{"model_service": model_serviceTimeStaticSubgraph, "config_params": config_params}),
    ]

    def get_routes(self):
        return self._routes

    def __init__(self):

        settings = {"debug": True}
        setup_swagger(self._routes,
                      swagger_url='/doc',
                      api_base_url='/',
                      description='',
                      api_version='1.0.0',
                      title='Journal API',
                      contact='name@domain',
                      schemes=['https'],
                      security_definitions={
                          'ApiKeyAuth': {
                              'type': 'apiKey',
                              'in': 'header',
                              'name': 'X-API-Key'
                          }
                      })
        super(Application, self).__init__(self._routes, **settings)


def main():
    App = Application()
    App.get_routes()
    App.listen(10089)
    tornado.ioloop.IOLoop.current().start()

if __name__ == "__main__":
    main()




import os
import sys
import ujson
from clickhouse_driver import Client
from graph_op import CHGraph
import pandas as pd


'''

graph_dir = "./config/tcpflow_flow.cfg.json"

#graph = CHGraph(graph_dir, client)

'''


def load_config():
    print("服务所有配置开始加载")
    curPath = os.path.abspath(os.path.dirname(__file__))
    rootPath = os.path.split(curPath)
    sys.path.append(rootPath[0])
    sys.path.append(rootPath[0]+"/"+rootPath[1])

    clickhouse_config_dir = rootPath[0]+"/"+rootPath[1]+"/config/"+"graph_config.json"

    with open(clickhouse_config_dir, 'r') as f:
        clickhouse_config = ujson.load(f)
   
    clickhouse_ip = clickhouse_config["ip"]
    print(clickhouse_ip)
    client = Client(clickhouse_ip)
    graph = CHGraph(client)
    print("服务所有配置加载结束")
    config_params = {
        "graph": graph
    }
    #config_params = {
    # 
    #}
    return config_params



import os
import sys
import json
from datetime import date,datetime
import pandas as pd
import numpy as np
import logging


LOG_FORMAT="%(asctime)s:%(levelname)s:%(name)s:%(funcName)s:%(message)s"
#logging.basicConfig(filename='./CHGraph.log', level=logging.INFO, format=LOG_FORMAT)
logging.basicConfig(filename='./CHGraph.log', level=logging.WARNING, format=LOG_FORMAT)
logger = logging.getLogger('CHGraph')


def tostr(obj):
    if type(obj) == list:
        return str(obj)
    elif type(obj) == np.ndarray:
        return np.array2string(obj, separator=',', threshold=1000000000, max_line_width=1000000000)


class CHGraph(object):

    def __init__(self, client):
        self.client = client
        logger.info('CHGraph Start')
        client.execute('create database if not exists graph_cfg')        
        client.execute('create table if not exists graph_cfg.name2cfg (graph_create_date Date,graph_name String,graph_status UInt32,graph_cfg String) engine = MergeTree(graph_create_date, (graph_name, graph_status), 8192);')

    #############################################
    #################### graph-wise operation
    #############################################
    # graph_name: graph name as string 
    # graph_cfg: graph cfg as json string
    def register_graph(self, graph_name, graph_cfg):        
        sql = "select * from graph_cfg.name2cfg where graph_name=\'"+graph_name+"\' and graph_status=1"
        res = self.client.execute(sql)
        if len(res) > 0:
            logger.warning("graph name ["+graph_name+"] already exists")
            return "graph name ["+graph_name+"] already exists"
        sql = "insert into graph_cfg.name2cfg (*) values (\'" + str(date.today()) + "\',\'" + graph_name + "\',1,\'" + graph_cfg + "\');"
        self.client.execute(sql)
        logger.info("register graph ["+graph_name+"] done")
        return "register graph ["+graph_name+"] done"


    def delete_graph(self, graph_name):
        sql = "select * from graph_cfg.name2cfg where graph_name=\'"+graph_name+"\' and graph_status=1"
        res = self.client.execute(sql)
        if len(res) == 0:
            logger.warning("graph name ["+graph_name+"] does not exist")
            return "graph name ["+graph_name+"] does not exist"
        sql_update = "alter table graph_cfg.name2cfg update graph_status=0 WHERE graph_name=\'"+graph_name+"\' and graph_status=1"
        self.client.execute(sql_update)
        logger.info("delete graph ["+graph_name+"] done")
        return "delete graph ["+graph_name+"] done"
        
        
    def show_graph(self):
        sql = "select graph_name from graph_cfg.name2cfg where graph_status=1"
        res = self.client.query_dataframe(sql)
        with pd.option_context('display.max_rows', None, 'display.max_columns', None):
            logger.info(res)
        return res

    def use_graph(self, graph_name):
        sql = "select * from graph_cfg.name2cfg where graph_name=\'" + graph_name + "\' and graph_status=1"
        res = self.client.execute(sql)
        if len(res) == 0:
            logger.warning("graph name [" + graph_name + "] does not exist")
            return
        else:
            self.graph_name = graph_name
            self.graph_cfg = json.loads(res[0][3])
            logger.info("use graph [" + graph_name + "] done")

    def describe_graph(self, graph_name):
        sql = "select * from graph_cfg.name2cfg where graph_name=\'"+graph_name+"\' and graph_status=1"
        res = self.client.execute(sql)
        if len(res) == 0:
            logger.warning("graph name ["+graph_name+"] does not exist")
            return "graph name ["+graph_name+"] does not exist"
        else:
            graph_cfg = json.loads(res[0][3])
            logger.info(graph_cfg)
            return graph_cfg

    def summary_graph(self, graph_name):
        sql = "select * from graph_cfg.name2cfg where graph_name=\'"+graph_name+"\' and graph_status=1"
        res = self.client.execute(sql)
        if len(res) == 0:
            logger.warning("graph name ["+graph_name+"] does not exist")
            return None
        graph_cfg = json.loads(res[0][3])
        edge_count = {}
        for name in graph_cfg['edges']:
            info = graph_cfg['edges'][name]
            sql = "select count(*) from "+info['db']+"."+info['table']
            res = self.client.execute(sql)
            logger.info("count of edge ["+name+"]: "+str(res[0][0]))
            edge_count[name] = res[0][0]
        vertex_count = {}
        for name in graph_cfg['vertexes']:
            info = graph_cfg['vertexes'][name]
            sql = "select count(*) from "+info['db']+"."+info['table']
            res = self.client.execute(sql)
            logger.info("count of vertex ["+name+"]: "+str(res[0][0]))
            vertex_count[name] = res[0][0]
        return (edge_count, vertex_count)


    #############################################
    #################### vertex/edge-wise operation: insert
    #############################################
    def insert_edge(self, edge_name, edge_schema, edge_data):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return "invalid edge: " + edge_name

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        if edge_info["src"] not in edge_schema or edge_info["dst"] not in edge_schema:
            logger.warning("require src field and dst field of edge")
            return "require src field and dst field of edge"

        sql = "insert into " + db_table + " (" + ",".join(edge_schema) + ") values " + ",".join([str(line) for line in edge_data])
        self.client.execute(sql)
        return "edge inserted success"


    def insert_vertex(self, vertex_name, vertex_schema, vertex_data): 
        
        if vertex_name not in self.graph_cfg["vertexes"]:
            logger.warning("invalid vertex: " + vertex_name)
            return "invalid vertex: " + vertex_name

        vertex_info = self.graph_cfg["vertexes"][vertex_name]
        db_table = vertex_info["db"] + "." + vertex_info["table"]

        if vertex_info["id"] not in vertex_schema:
            logger.warning("require id field of vertex")
            return "require id field of vertex"

        sql = "insert into " + db_table + " (" + ",".join(vertex_schema) + ") values " + ",".join([str(line) for line in vertex_data])
        self.client.execute(sql)
        return "vertex inserted success"


    #############################################
    #################### vertex/edge-wise operation: search
    #############################################
    ## one_hop return edges
    ## one_hop does not require type(src) == type(dst)
    def one_hop(self,
                start_vertex_list,
                direction,
                edge_name,
                edge_con_list, 
                target_field_list,
                end_vertex_con_list=None):
        
        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return

        edge_info = self.graph_cfg["edges"][edge_name]

        db_table = edge_info["db"] + "." + edge_info["table"]

        if direction == "forward":
            start_vertex_list_con = edge_info["src"] + " in " + tostr(start_vertex_list)
        elif direction == "backward":
            start_vertex_list_con = edge_info["dst"] + " in " + tostr(start_vertex_list)
        elif direction == "bidirectional":
            start_vertex_list_con = "(" + edge_info["src"] + " in " + tostr(start_vertex_list) + " or " + edge_info["dst"] + " in " + tostr(start_vertex_list) + ")"
        else:
            logger.warning("invalid direction")
            return "invalid direction"
        #print(start_vertex_list_con)

        edge_con_list_con = ""
        if edge_con_list is not None and len(edge_con_list) > 0:
            edge_con_list_con = " and " + " and ".join(edge_con_list)
        #print(edge_con_list_con)

        target_field_list_con = edge_info["src"] + "," + edge_info["dst"]
        if target_field_list is None:
            target_field_list_con += "," + ",".join(edge_info["fields"])
        elif type(target_field_list) == list and len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)
        elif type(target_field_list) == str and target_field_list == "src":
            target_field_list_con = "DISTINCT " + edge_info["src"]
        elif type(target_field_list) == str and target_field_list == "dst":
            target_field_list_con = "DISTINCT " + edge_info["dst"]

        sql = "select " + target_field_list_con + " from " + db_table + " where " + \
                  start_vertex_list_con + edge_con_list_con
        logger.info("sql: "+sql)
        #res = self.client.execute(sql)
        res = self.client.query_dataframe(sql)

        return res

    
    ## multi_hop return edges
    ## multi_hop requires type(src) == type(dst)
    def multi_hop(self,
                  step,
                  start_vertex_list,
                  direction,
                  edge_name,
                  edge_con_list, 
                  target_field_list,
                  only_last_step,
                  plus_last_vertexes=False,
                  end_vertex_con_list=None):

        multi_res = []
        
        multi_step_start_vertex_list = start_vertex_list

        for i in range(step):

            res = self.one_hop(multi_step_start_vertex_list,
                               direction,
                               edge_name,
                               edge_con_list,
                               target_field_list)

            if res.shape[0] == 0:
                logger.warning("multi-hop terminates at "+str(i+1))
                break
            
            multi_res.append(res)
            
            if i == step - 1 and not plus_last_vertexes:
                continue

            # TODO: may have performance issue
            edge_info = self.graph_cfg["edges"][edge_name]
            if direction == "forward":
                multi_step_start_vertex_list = np.unique(res[edge_info['dst']].values)
            elif direction == "backward":
                multi_step_start_vertex_list = np.unique(res[edge_info['src']].values)
            elif direction == "bidirectional":
                logger.warning("bidirectional not implemented")
                return
            else:
                logger.warning("invalid direction")
                return "invalid direction"

        if only_last_step:
            if plus_last_vertexes:
                return (multi_res[-1], multi_step_start_vertex_list)
            else:
                return multi_res[-1]
        else:
            return multi_res


    # multi_hop_common_vertexes return common vertexes
    def multi_hop_common_vertexes(self,
                                  step,
                                  start_vertex_list,
                                  direction,
                                  edge_name,
                                  edge_con_list):
        
        end_set_list = []
        for i in range(len(start_vertex_list)):
            (res, vertex) = self.multi_hop(step, 
                                           [start_vertex_list[i]], 
                                           direction,
                                           edge_name,
                                           edge_con_list,
                                           [],
                                           True, 
                                           True)
            if direction == "forward":
                end_set_list.append(set(vertex))
            elif direction == "backward":
                end_set_list.append(set(vertex))
            elif direction == "bidirectional":
                logger.warning("bidirectional not implemented")
                return
            else:
                logger.warning("invalid direction")
                return
        
        intersect = end_set_list[0]
        for i in range(1, len(end_set_list)):
            intersect &= end_set_list[i]

        return list(intersect)
        

    ## one_hop_multi_edge return edges
    ## one_hop_multi_edge does not require type(src) == type(dst)
    def one_hop_multi_edge(self,
                           start_vertex_list,
                           direction,
                           edge_name_list,
                           edge_con_list_list,
                           target_field_list,
                           end_vertex_con_list=None):
        
        res_list = []

        for edge_name in edge_name_list:

            if edge_name not in self.graph_cfg["edges"]:
                logger.warning("invalid edge: " + edge_name)
                return "invalid edge: " + edge_name
            
        for i in range(len(edge_name_list)):

            edge_name = edge_name_list[i]
            edge_con_list = edge_con_list_list[i]

            res = self.one_hop(start_vertex_list, direction, edge_name, edge_con_list, target_field_list)

            res_list.append(res)

        return res_list


    ## multi_hop_multi_edge return edges
    ## multi_hop_multi_edge requires type(src) == type(dst)
    def multi_hop_multi_edge(self,
                             step,
                             start_vertex_list,
                             direction,
                             edge_name_list,
                             edge_con_list_list,
                             target_field_list,
                             only_last_step,
                             plus_last_vertexes=False):

        multi_res = []

        multi_step_start_vertex_list = start_vertex_list

        for i in range(step):

            res = self.one_hop_multi_edge(multi_step_start_vertex_list,
                                          direction,
                                          edge_name_list,
                                          edge_con_list_list,
                                          target_field_list)

            multi_res.append(res)

            #for item in multi_res:
            #    print(item)

            if sum([res_elem.shape[0] for res_elem in res]) == 0:
                logger.warning("multi-hop terminates at " + str(i + 1))
                multi_step_start_vertex_list = []
                break

            if i == step - 1 and not plus_last_vertexes:
                continue

            # TODO: may have performance issue
            if direction == "forward":
                tmp_list = []
                for ii in range(len(edge_name_list)):
                    if res[ii].shape[0] == 0:
                        continue
                    edge_name = edge_name_list[ii]
                    edge_info = self.graph_cfg["edges"][edge_name]
                    tmp_list.append(res[ii][edge_info['dst']].values)
                multi_step_start_vertex_list = np.unique(np.concatenate(tmp_list))
            elif direction == "backward":
                tmp_list = []
                for ii in range(len(edge_name_list)):
                    if res[ii].shape[0] == 0:
                        continue
                    edge_name = edge_name_list[ii]
                    edge_info = self.graph_cfg["edges"][edge_name]
                    tmp_list.append(res[ii][edge_info['src']].values)
                multi_step_start_vertex_list = np.unique(np.concatenate(tmp_list))
            elif direction == "bidirectional":
                logger.warning("bidirectional not implemented")
                return "bidirectional not implemented"
            else:
                logger.warning("invalid direction")
                return "invalid direction"

        if only_last_step:
            if plus_last_vertexes:
                return (multi_res[-1], multi_step_start_vertex_list)
            else:
                return multi_res[-1]
        else:
            return multi_res


    # multi_hop_multi_edge_common_vertexes return common vertexes
    def multi_hop_multi_edge_common_vertexes(self,
                                             step,
                                             start_vertex_list,
                                             direction,
                                             edge_name_list,
                                             edge_con_list_list):
        
        end_set_list = []
        for i in range(len(start_vertex_list)):
            (res, vertex) = self.multi_hop_multi_edge(step, 
                                                      [start_vertex_list[i]], 
                                                      direction,
                                                      edge_name_list,
                                                      edge_con_list_list,
                                                      [],
                                                      True,
                                                      True)
            if direction == "forward":
                end_set_list.append(set(vertex))
            elif direction == "backward":
                end_set_list.append(set(vertex))
            elif direction == "bidirectional":
                logger.warning("bidirectional not implemented")
                return "bidirectional not implemented"
            else:
                logger.warning("invalid direction")
                return "invalid direction"
        
        intersect = end_set_list[0]
        for i in range(1, len(end_set_list)):
            intersect &= end_set_list[i]

        return list(intersect)
        

    # match_edge returns edges which satisfy constraints
    def match_edge(self,
                   edge_name,
                   edge_con_list,
                   target_field_list,
                   data_type="list"):
        
        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return "invalid edge: " + edge_name

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        edge_con_list_con = ""
        if edge_con_list is not None and len(edge_con_list) > 0:
            edge_con_list_con = " where " + " and ".join(edge_con_list)

        target_field_list_con = edge_info["src"] + "," + edge_info["dst"]
        if target_field_list is None:
            target_field_list_con += "," + ",".join(edge_info["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)

        sql = "select " + target_field_list_con + " from " + db_table + edge_con_list_con
        logger.info("sql: "+sql)
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql)

        return res 


    # match_vertex returns vertexes which satisfy constraints
    def match_vertex(self,
                     vertex_name,
                     vertex_con_list,
                     target_field_list,
                     data_type="list"):
        
        if vertex_name not in self.graph_cfg["vertexes"]:
            logger.warning("invalid vertex: " + vertex_name)
            return "invalid vertex: " + vertex_name

        vertex_info = self.graph_cfg["vertexes"][vertex_name]
        db_table = vertex_info["db"] + "." + vertex_info["table"]

        vertex_con_list_con = ""
        if vertex_con_list is not None and len(vertex_con_list) > 0:
            vertex_con_list_con = " where " + " and ".join(vertex_con_list)

        target_field_list_con = vertex_info["id"]
        if target_field_list is None:
            target_field_list_con += "," + ",".join(vertex_info["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)

        sql = "select " + target_field_list_con + " from " + db_table + vertex_con_list_con
        logger.info("sql: "+sql)
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql)

        return res 

    # def find_path_multi_edge(self,
    #                          start_vertex_list,
    #                          end_vertex_list,
    #                          edge_name_list,
    #                          edge_con_list_list,
    #                          target_field_list,
    #                          step_limit):
    #     if step_limit > 2:
    #         logger.warning("finding path longer than 2 not implemented")
    #         res2 = self.find_path_multi_edge_s2(start_vertex_list,
    #                                             end_vertex_list,
    #                                             edge_name_list,
    #                                             edge_con_list_list,
    #                                             target_field_list)
    #
    #         res1 = self.find_path_multi_edge_s1(start_vertex_list,
    #                                             end_vertex_list,
    #                                             edge_name_list,
    #                                             edge_con_list_list,
    #                                             target_field_list)
    #         return [res1, res2]
    #
    #     if step_limit > 1:
    #         res2 = self.find_path_multi_edge_s2(start_vertex_list,
    #                                             end_vertex_list,
    #                                             edge_name_list,
    #                                             edge_con_list_list,
    #                                             target_field_list)
    #
    #         res1 = self.find_path_multi_edge_s1(start_vertex_list,
    #                                             end_vertex_list,
    #                                             edge_name_list,
    #                                             edge_con_list_list,
    #                                             target_field_list)
    #         return [res1, res2]
    #
    #     elif step_limit > 0:
    #         res1 = self.find_path_multi_edge_s1(start_vertex_list,
    #                                             end_vertex_list,
    #                                             edge_name_list,
    #                                             edge_con_list_list,
    #                                             target_field_list)
    #         return [res1]
    #
    # def find_path_multi_edge_s1(self,
    #                             start_vertex_list,
    #                             end_vertex_list,
    #                             edge_name_list,
    #                             edge_con_list_list,
    #                             target_field_list):
    #
    #     res_list = []
    #
    #     for i in range(len(edge_name_list)):
    #         edge_name = edge_name_list[i]
    #         edge_con_list = edge_con_list_list[i]
    #         edge_info = self.graph_cfg["edges"][edge_name]
    #         start_vertex_list_con = edge_info["src"] + " in " + tostr(start_vertex_list)
    #         end_vertex_list_con = edge_info["dst"] + " in " + tostr(end_vertex_list)
    #         edge_con_list.append(start_vertex_list_con)
    #         edge_con_list.append(end_vertex_list_con)
    #         res = self.match_edge(edge_name, edge_con_list, target_field_list, data_type="df")
    #         res_list.append(res)
    #
    #     return res_list
    #
    # def find_path_multi_edge_s2(self,
    #                             start_vertex_list,
    #                             end_vertex_list,
    #                             edge_name_list,
    #                             edge_con_list_list,
    #                             target_field_list):
    #
    #     (f_edges, f_vertices) = self.multi_hop_multi_edge(1, start_vertex_list, "forward",
    #                                                       edge_name_list, edge_con_list_list,
    #                                                       target_field_list, True, True)
    #
    #     (b_edges, b_vertices) = self.multi_hop_multi_edge(1, end_vertex_list, "backward",
    #                                                       edge_name_list, edge_con_list_list,
    #                                                       target_field_list, True, True)
    #
    #     transit_vertices = np.intersect1d(f_vertices, b_vertices, assume_unique=True)
    #     f_edges_final = []
    #     for df in f_edges:
    #         if type(df) == list:
    #            if len(df) != 0:
    #                df = pd.DataFrame(df)
    #                f_edges_final.append(df[df.iloc[:, 1].isin(transit_vertices)])
    #            else:
    #                f_edges_final.append(pd.DataFrame([]))
    #     b_edges_final = []
    #     for df in b_edges:
    #         if type(df) == list:
    #            if len(df) != 0:
    #                df = pd.DataFrame(df)
    #                b_edges_final.append(df[df.iloc[:, 0].isin(transit_vertices)])
    #            else:
    #                b_edges_final.append(pd.DataFrame([]))
    #
    #     return [f_edges_final, b_edges_final]

    def find_path_multi_edge(self,
                             start_vertex_list,
                             end_vertex_list,
                             edge_name_list,
                             edge_con_list_list,
                             target_field_list,
                             step_limit):

        if step_limit > 3:
            # logger.warning("finding path longer than 2 not implemented")

            res4 = self.find_path_multi_edge_s4(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)

            res3 = self.find_path_multi_edge_s3(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)

            res2 = self.find_path_multi_edge_s2(start_vertex_list,
                                                 end_vertex_list,
                                                 edge_name_list,
                                                 edge_con_list_list,
                                                 target_field_list)
            
            res1 = self.find_path_multi_edge_s1(start_vertex_list,
                                                 end_vertex_list,
                                                 edge_name_list,
                                                 edge_con_list_list,
                                                 target_field_list)
            
            return [res1, res2, res3, res4]

        if step_limit > 2:
            res3 = self.find_path_multi_edge_s3(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)
            # logger.warning("finding path longer than 2 not implemented")
            res2 = self.find_path_multi_edge_s2(start_vertex_list,
                                                 end_vertex_list,
                                                 edge_name_list,
                                                 edge_con_list_list,
                                                 target_field_list)
 
            res1 = self.find_path_multi_edge_s1(start_vertex_list,
                                                 end_vertex_list,
                                                 edge_name_list,
                                                 edge_con_list_list,
                                                 target_field_list)
            return [res1, res2, res3]

        if step_limit > 1:
            res2 = self.find_path_multi_edge_s2(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)

            res1 = self.find_path_multi_edge_s1(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)
            return [res1, res2]

        if step_limit > 0:
            res1 = self.find_path_multi_edge_s1(start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                target_field_list)
            return [res1]

    # return path of length 1 (list of df)
    def find_path_multi_edge_s1(self,
                                start_vertex_list,
                                end_vertex_list,
                                edge_name_list,
                                edge_con_list_list,
                                target_field_list):

        res_list = []

        for i in range(len(edge_name_list)):
            edge_name = edge_name_list[i]
            edge_con_list = edge_con_list_list[i]
            edge_info = self.graph_cfg["edges"][edge_name]
            start_vertex_list_con = edge_info["src"] + " in " + tostr(start_vertex_list)
            end_vertex_list_con = edge_info["dst"] + " in " + tostr(end_vertex_list)
            edge_con_list.append(start_vertex_list_con)
            edge_con_list.append(end_vertex_list_con)
            res = self.match_edge(edge_name, edge_con_list, target_field_list, data_type="df")
            res_list.append(res)

        return res_list

        # return path of length 2 (list of list of df)

    def find_path_multi_edge_s2(self,
                                start_vertex_list,
                                end_vertex_list,
                                edge_name_list,
                                edge_con_list_list,
                                target_field_list):


        (f_edges, f_vertices) = self.multi_hop_multi_edge(1, start_vertex_list, "forward",
                                                          edge_name_list, edge_con_list_list,
                                                          target_field_list, True, True)

        (b_edges, b_vertices) = self.multi_hop_multi_edge(1, end_vertex_list, "backward",
                                                          edge_name_list, edge_con_list_list,
                                                          target_field_list, True, True)

        transit_vertices = np.intersect1d(f_vertices, b_vertices, assume_unique=True)

        if len(transit_vertices) == 0:
            return [[pd.DataFrame([])]*len(edge_name_list), [pd.DataFrame([])]*len(edge_name_list)]

        f_edges_final = []
        for df in f_edges:
            if len(df) != 0:
                f_edges_final.append(df[df.iloc[:, 1].isin(transit_vertices)])
            else:
                f_edges_final.append(pd.DataFrame([]))
        b_edges_final = []
        for df in b_edges:
            if len(df) != 0:
                b_edges_final.append(df[df.iloc[:, 0].isin(transit_vertices)])
            else:
                b_edges_final.append(pd.DataFrame([]))
        return [f_edges_final, b_edges_final]


    def find_path_multi_edge_s3(self,
                                start_vertex_list,
                                end_vertex_list,
                                edge_name_list,
                                edge_con_list_list,
                                target_field_list):

        # f_vertices is ndarray  & f_edges is list of dataframe
        (first_edges, f_vertices) = self.multi_hop_multi_edge(1, start_vertex_list, "forward",
                                                              edge_name_list, edge_con_list_list,
                                                              target_field_list, True, True)
        # f_edges_final, b_edges_final is list of datdaframe
        [second_edges, third_edges_final] = self.find_path_multi_edge_s2(f_vertices.tolist(), end_vertex_list,
                                                                         edge_name_list,
                                                                         edge_con_list_list, target_field_list)

        # for  backward
        if second_edges == "":
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list)]
        tmp_list = []
        for ii in range(len(edge_name_list)):
            if second_edges[ii].shape[0] == 0:
                continue
            edge_name = edge_name_list[ii]
            edge_info = self.graph_cfg["edges"][edge_name]
            tmp_list.append(second_edges[ii][edge_info['src']].values)
        if tmp_list==[]:
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list),
                    [pd.DataFrame([])] * len(edge_name_list)]
        multi_step_start_vertex_list = np.unique(np.concatenate(tmp_list))
        transit_vertices = np.intersect1d(f_vertices, multi_step_start_vertex_list, assume_unique=True)
        if len(transit_vertices) == 0:
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list)]
        first_edges_final = []
        for df in first_edges:
            if df.empty:
                first_edges_final.append(pd.DataFrame([]))
                continue
            first_edges_final.append(df[df.iloc[:, 1].isin(transit_vertices)])
        second_edges_final = []
        for df in second_edges:
            if df.empty:
                second_edges_final.append(pd.DataFrame([]))
                continue
            second_edges_final.append(df[df.iloc[:, 0].isin(transit_vertices)])

        return [first_edges_final, second_edges_final, third_edges_final]


    def find_path_multi_edge_s4(self,
                                start_vertex_list,
                                end_vertex_list,
                                edge_name_list,
                                edge_con_list_list,
                                target_field_list):

        # f_vertices is ndarray  & f_edges is list of dataframe
        (first_edges, f_vertices) = self.multi_hop_multi_edge(1, start_vertex_list, "forward",
                                                              edge_name_list, edge_con_list_list,
                                                              target_field_list, True, True)
        # f_edges_final, b_edges_final is list of datdaframe
        [second_edges, third_edges_final, fourth_edges_final] = self.find_path_multi_edge_s3(f_vertices.tolist(),
                                                                                             end_vertex_list,
                                                                                             edge_name_list,
                                                                                             edge_con_list_list,
                                                                                             target_field_list)
        if second_edges == "":
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list),[pd.DataFrame([])] * len(edge_name_list)]
        tmp_list = []
        for ii in range(len(edge_name_list)):
            if second_edges[ii].shape[0] == 0:
                continue
            edge_name = edge_name_list[ii]
            edge_info = self.graph_cfg["edges"][edge_name]
            tmp_list.append(second_edges[ii][edge_info['src']].values)
        if tmp_list==[]:
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list),
                    [pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list)]
        multi_step_start_vertex_list = np.unique(np.concatenate(tmp_list))
        transit_vertices = np.intersect1d(f_vertices, multi_step_start_vertex_list, assume_unique=True)
        if len(transit_vertices) == 0:
            return [[pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list), [pd.DataFrame([])] * len(edge_name_list),[pd.DataFrame([])] * len(edge_name_list)]
        first_edges_final = []
        for df in first_edges:
            if df.empty:
                first_edges_final.append(pd.DataFrame([]))
                continue
            first_edges_final.append(df[df.iloc[:, 1].isin(transit_vertices)])
        second_edges_final = []
        for df in second_edges:
            if df.empty:
                second_edges_final.append(pd.DataFrame([]))
                continue
            second_edges_final.append(df[df.iloc[:, 0].isin(transit_vertices)])

        return [first_edges_final, second_edges_final, third_edges_final, fourth_edges_final]




        #############################################
        #################### subgraph
        #############################################

        # subgraph_name is recommended to be unique (ex. plus timestamp)

    def create_subgraph(self, subgraph_name):

        self.client.execute("create database if not exists " + subgraph_name)
        try:
            for name in self.graph_cfg["edges"]:
                info = self.graph_cfg["edges"][name]
                new_name = subgraph_name + "." + info['table']
                old_name = info['db'] + "." + info['table']
                self.client.execute("create table " + new_name + " as " + old_name)
            for name in self.graph_cfg["vertexes"]:
                info = self.graph_cfg["vertexes"][name]
                new_name = subgraph_name + "." + info['table']
                old_name = info['db'] + "." + info['table']
                self.client.execute("create table " + new_name + " as " + old_name)
        except Exception as e:
            print(e)
            return "subgraph is already exist or something wrong"
        import copy
        subgraph_cfg = copy.deepcopy(self.graph_cfg)
        for name in subgraph_cfg["edges"]:
            subgraph_cfg["edges"][name]["db"] = subgraph_name
        for name in subgraph_cfg["vertexes"]:
            subgraph_cfg["vertexes"][name]["db"] = subgraph_name
        subgraph_cfg["subgraph"] = self.graph_name
        return self.register_graph(subgraph_name, json.dumps(subgraph_cfg))

    def update_subgraph_by_multi_hop_multi_edge(self,
                                                subgraph_name,
                                                step,
                                                start_vertex_list,
                                                direction,
                                                edge_name_list,
                                                edge_con_list_list):

        multi_res = self.multi_hop_multi_edge(step, start_vertex_list, direction,
                                              edge_name_list, edge_con_list_list, None, False, False)


        try:
            for single_res in multi_res:
                for i in range(len(edge_name_list)):
                    if single_res[i].shape[0] == 0:
                        continue
                    edge_info = self.graph_cfg["edges"][edge_name_list[i]]
                    db_table = subgraph_name + "." + edge_info["table"]
                    target_field_list_con = edge_info["src"] + "," + edge_info["dst"] + "," + ",".join(
                        edge_info["fields"])
                    # TODO: list hurts performance and increases memory cost
                    # datetime64[ns] patch
                    # col_list = [list(single_res[i][col].values) for col in single_res[i].columns]
                    col_list = [list(pd.to_datetime(single_res[i][col])) if str(
                        single_res[i][col].dtype) == "datetime64[ns]" else list(single_res[i][col].values) for col in
                                single_res[i].columns]
                    self.client.execute("insert into " + db_table + " (" + target_field_list_con + ") values", col_list,
                                        columnar=True)

                    src_info = self.graph_cfg["vertexes"][edge_info["src_type"]]
                    src_db_table_new = subgraph_name + "." + src_info["table"]
                    src_db_table_old = src_info["db"] + "." + src_info["table"]
                    src_target_field_list_con = src_info["id"] + "," + ",".join(src_info["fields"])
                    self.client.execute(
                        "insert into " + src_db_table_new + " (" + src_target_field_list_con + ") select " + src_target_field_list_con + " from " + src_db_table_old + " where " +
                        src_info["id"] + " in " + tostr(col_list[0]))
                    dst_info = self.graph_cfg["vertexes"][edge_info["dst_type"]]
                    dst_db_table_new = subgraph_name + "." + dst_info["table"]
                    dst_db_table_old = dst_info["db"] + "." + dst_info["table"]
                    dst_target_field_list_con = dst_info["id"] + "," + ",".join(dst_info["fields"])
                    self.client.execute(
                        "insert into " + dst_db_table_new + " (" + dst_target_field_list_con + ") select " + dst_target_field_list_con + " from " + dst_db_table_old + " where " +
                        dst_info["id"] + " in " + tostr(col_list[1]))
            return "The subgraph is update success"
        except Exception as e:
            print(e)
            return "The subgraph is update failed"
       

    def update_subgraph_by_match_edge(self,
                                      subgraph_name,
                                      edge_name,
                                      edge_con_list):

        res = self.match_edge(edge_name, edge_con_list, None, data_type='df')
        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = subgraph_name + "." + edge_info["table"]
        target_field_list_con = edge_info["src"] + "," + edge_info["dst"] + "," + ",".join(edge_info["fields"])
        # TODO: list hurts performance and increases memory cost
        col_list = [list(pd.to_datetime(res[col])) if str(res[col].dtype) == "datetime64[ns]" else list(res[col].values) for col in res.columns]
        try:
            self.client.execute("insert into "+db_table+" ("+target_field_list_con+") values", col_list, columnar=True)

            src_info = self.graph_cfg["vertexes"][edge_info["src_type"]]
            src_db_table_new = subgraph_name + "." + src_info["table"]
            src_db_table_old = src_info["db"] + "." + src_info["table"]
            src_target_field_list_con = src_info["id"] + "," + ",".join(src_info["fields"])
            self.client.execute(
                "insert into " + src_db_table_new + " (" + src_target_field_list_con + ") select " + src_target_field_list_con + " from " + src_db_table_old + " where " +
                src_info["id"] + " in " + tostr(col_list[0]))
            dst_info = self.graph_cfg["vertexes"][edge_info["dst_type"]]
            dst_db_table_new = subgraph_name + "." + dst_info["table"]
            dst_db_table_old = dst_info["db"] + "." + dst_info["table"]
            dst_target_field_list_con = dst_info["id"] + "," + ",".join(dst_info["fields"])
            self.client.execute(
                "insert into " + dst_db_table_new + " (" + dst_target_field_list_con + ") select " + dst_target_field_list_con + " from " + dst_db_table_old + " where " +
                dst_info["id"] + " in " + tostr(col_list[1]))

            return "The subgraph update success"
        except Exception as e:
            print(e)
            return "The subgraph update failed"

    def update_subgraph_by_find_path_multi_edge(self,
                                                subgraph_name,
                                                start_vertex_list,
                                                end_vertex_list,
                                                edge_name_list,
                                                edge_con_list_list,
                                                step_limit):

        res = self.find_path_multi_edge(start_vertex_list, end_vertex_list,
                                        edge_name_list, edge_con_list_list,
                                        None, step_limit)

        for i in range(len(res)):
            if i == 0:  # path of length 1
                for k in range(len(edge_name_list)):
                    if res[i][k].shape[0] == 0:
                        continue
                    edge_info = self.graph_cfg["edges"][edge_name_list[k]]
                    db_table = subgraph_name + "." + edge_info["table"]
                    target_field_list_con = edge_info["src"] + "," + edge_info["dst"] + "," + ",".join(
                        edge_info["fields"])
                    # TODO: list hurts performance and increases memory cost
                    # datetime64[ns] patch
                    # col_list = [list(res[i][k][col].values) for col in res[i][k].columns]
                    col_list = [
                        list(pd.to_datetime(res[i][k][col])) if str(res[i][k][col].dtype) == "datetime64[ns]" else list(
                            res[i][k][col].values) for col in res[i][k].columns]
                    self.client.execute("insert into " + db_table + " (" + target_field_list_con + ") values", col_list,
                                        columnar=True)
                    src_info = self.graph_cfg["vertexes"][edge_info["src_type"]]
                    src_db_table_new = subgraph_name + "." + src_info["table"]
                    src_db_table_old = src_info["db"] + "." + src_info["table"]
                    src_target_field_list_con = src_info["id"] + "," + ",".join(src_info["fields"])
                    self.client.execute(
                        "insert into " + src_db_table_new + " (" + src_target_field_list_con + ") select " + src_target_field_list_con + " from " + src_db_table_old + " where " +
                        src_info["id"] + " in " + tostr(col_list[0]))
                    dst_info = self.graph_cfg["vertexes"][edge_info["dst_type"]]
                    dst_db_table_new = subgraph_name + "." + dst_info["table"]
                    dst_db_table_old = dst_info["db"] + "." + dst_info["table"]
                    dst_target_field_list_con = dst_info["id"] + "," + ",".join(dst_info["fields"])
                    self.client.execute(
                        "insert into " + dst_db_table_new + " (" + dst_target_field_list_con + ") select " + dst_target_field_list_con + " from " + dst_db_table_old + " where " +
                        dst_info["id"] + " in " + tostr(col_list[1]))

            else:  # path of length >=2
                for j in range(len(res[i])):
                    for k in range(len(edge_name_list)):
                        if res[i][j][k].shape[0] == 0:
                            continue
                        edge_info = self.graph_cfg["edges"][edge_name_list[k]]
                        db_table = subgraph_name + "." + edge_info["table"]
                        target_field_list_con = edge_info["src"] + "," + edge_info["dst"] + "," + ",".join(
                            edge_info["fields"])
                        # TODO: list hurts performance and increases memory cost
                        # datetime64[ns] patch
                        # col_list = [list(res[i][j][k][col].values) for col in res[i][j][k].columns]
                        col_list = [list(pd.to_datetime(res[i][j][k][col])) if str(
                            res[i][j][k][col].dtype) == "datetime64[ns]" else list(res[i][j][k][col].values) for col in
                                    res[i][j][k].columns]
                        self.client.execute("insert into " + db_table + " (" + target_field_list_con + ") values",
                                            col_list, columnar=True)
                        src_info = self.graph_cfg["vertexes"][edge_info["src_type"]]
                        src_db_table_new = subgraph_name + "." + src_info["table"]
                        src_db_table_old = src_info["db"] + "." + src_info["table"]
                        src_target_field_list_con = src_info["id"] + "," + ",".join(src_info["fields"])
                        self.client.execute(
                            "insert into " + src_db_table_new + " (" + src_target_field_list_con + ") select " + src_target_field_list_con + " from " + src_db_table_old + " where " +
                            src_info["id"] + " in " + tostr(col_list[0]))
                        dst_info = self.graph_cfg["vertexes"][edge_info["dst_type"]]
                        dst_db_table_new = subgraph_name + "." + dst_info["table"]
                        dst_db_table_old = dst_info["db"] + "." + dst_info["table"]
                        dst_target_field_list_con = dst_info["id"] + "," + ",".join(dst_info["fields"])
                        self.client.execute(
                            "insert into " + dst_db_table_new + " (" + dst_target_field_list_con + ") select " + dst_target_field_list_con + " from " + dst_db_table_old + " where " +
                            dst_info["id"] + " in " + tostr(col_list[1]))

    def destroy_subgraph(self, subgraph_name):

        graph_name = self.graph_name
        self.use_graph(subgraph_name)
        if "subgraph" not in self.graph_cfg or self.graph_cfg["subgraph"] != graph_name:
            logger.warning("not subgraph or graph-subgraph unmatched")
            self.use_graph(graph_name)
            return "not subgraph or graph-subgraph unmatched"
        self.use_graph(graph_name)

        self.delete_graph(subgraph_name)
        self.client.execute("drop database if exists " + subgraph_name)
        return "destroy subgraph success"

    #############################################
    #################### vertex/edge-wise operation: metric
    #############################################
    ## metric_indegree returns indegree of vertices (df)
    ## vertices without in-edge will be ignored
    def metric_indegree(self, edge_name, if_sort=False, topk=-1):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return "invalid edge: " + edge_name

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        qualifier = ""
        if if_sort:
            qualifier += " order by indegree desc"
        if topk > 0:
            qualifier += " limit " + str(topk)

        sql = "select " + edge_info["dst"] + ", count(*) as indegree from " + db_table + " group by " + edge_info[
            "dst"] + qualifier
        logger.info("sql: " + sql)
        res = self.client.query_dataframe(sql)
        return res.to_dict()

    ## metric_outdegree returns outdegree of vertices (df)
    ## vertices without out-edge will be ignored
    def metric_outdegree(self, edge_name, if_sort=False, topk=-1):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return "invalid edge: " + edge_name

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        qualifier = ""
        if if_sort:
            qualifier += " order by outdegree desc"
        if topk > 0:
            qualifier += " limit " + str(topk)

        sql = "select " + edge_info["src"] + ", count(*) as outdegree from " + db_table + " group by " + edge_info[
            "src"] + qualifier
        logger.info("sql: " + sql)
        res = self.client.query_dataframe(sql)
        return res.to_dict()

    ## metric_degree returns degree of vertices (df)
    ## vertices without edge will be ignored
    def metric_degree(self, edge_name, if_sort=False, topk=-1):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        qualifier = ""
        if if_sort:
            qualifier += " order by degree desc"
        if topk > 0:
            qualifier += " limit " + str(topk)

        sql_o = "select " + edge_info["src"] + " as vid, count(*) as degree from " + db_table + " group by " + \
                edge_info["src"]
        sql_i = "select " + edge_info["dst"] + " as vid, count(*) as degree from " + db_table + " group by " + \
                edge_info["dst"]
        sql = "select vid, sum(degree) as degree from (" + sql_o + " union all " + sql_i + ") group by vid" + qualifier
        logger.info("sql: " + sql)
        res = self.client.query_dataframe(sql)
        return res.to_dict()

    ## TODO
    ## metric_pagerank returns pagerank of vertices
    ## vertices without edge will be ignored
    def metric_pagerank(self, edge_name, d=0.85, num_iter=10, if_sort=False, topk=-1):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        qualifier = ""
        if if_sort:
            qualifier += " order by outdegree desc"
        if topk > 0:
            qualifier += " limit " + str(topk)

        sql = "select " + edge_info["src"] + "," + edge_info["dst"] + ",outdegree from (select " + edge_info[
            "src"] + ",count(*) as outdegree from " + db_table + " group by " + edge_info[
                  "src"] + ") as table_od join " + db_table + " using " + edge_info["src"] + qualifier
        print(sql)
        res = self.client.query_dataframe(sql)
        return res.to_dict()

    def vertex_match_property(self, vertex_id_list, vertex_name, vertex_con_list, target_field_list, data_type="list"):
        propertys = self.graph_cfg["vertexes"]
        table_property = propertys[vertex_name]
        db_table = table_property["db"] + "." + table_property["table"]
        target_field_list_con = table_property["id"]
        if target_field_list is None or len(target_field_list) == 0:
            target_field_list_con += "," + ",".join(table_property["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)
        vertex_con_list_con = " where " + table_property["id"] + " in " + tostr(vertex_id_list)
        if vertex_con_list is not None and len(vertex_con_list) > 0:
            vertex_con_list_con += " and " + " and ".join(vertex_con_list)
        sql = "select DISTINCT " + target_field_list_con + " from " + db_table + vertex_con_list_con
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql).values.tolist()
        return res

    def edge_match_property(self, start_vertex_list, end_vertex_list, edge_name, edge_con_list, target_field_list,
                            data_type="list"):
        propertys = self.graph_cfg["edges"]
        table_property = propertys[edge_name]
        db_table = table_property["db"] + "." + table_property["table"]
        target_field_list_con = table_property["src"] + "," + table_property["dst"]
        if target_field_list is None or len(target_field_list)==0:
            target_field_list_con += "," + ",".join(table_property["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)
        edge_con_list_con = " where " + table_property["src"] + " in " + tostr(start_vertex_list) + " and " \
                            + table_property["dst"] + " in " + tostr(end_vertex_list)
        if edge_con_list is not None and len(edge_con_list) > 0:
            edge_con_list_con += " and " + " and ".join(edge_con_list)
        sql = "select DISTINCT " + target_field_list_con + " from " + db_table + edge_con_list_con
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql).values.tolist()
        return res

    # query_vertex returns distinct vertexes which satisfy constraints (df)
    def query_vertexes(self,
                     vertex_name,
                     vertex_con_list,
                     target_field_list,
                     data_type="list"):

        if vertex_name not in self.graph_cfg["vertexes"]:
            logger.warning("invalid vertex: " + vertex_name)
            return

        vertex_info = self.graph_cfg["vertexes"][vertex_name]
        db_table = vertex_info["db"] + "." + vertex_info["table"]

        vertex_con_list_con = ""
        if vertex_con_list is not None and len(vertex_con_list) > 0:
            vertex_con_list_con = " where " + " and ".join(vertex_con_list)

        target_field_list_con = vertex_info["id"]
        if target_field_list is None or len(target_field_list) == 0:
            target_field_list_con += "," + ",".join(vertex_info["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)

        sql = "select distinct " + target_field_list_con + " from " + db_table + vertex_con_list_con
        logger.info("sql: " + sql)
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql).values.tolist()

        return res

    # query_edge returns distinct edges which satisfy constraints (df)
    def query_edges(self,
                   edge_name,
                   edge_con_list,
                   target_field_list,
                   data_type="list"):

        if edge_name not in self.graph_cfg["edges"]:
            logger.warning("invalid edge: " + edge_name)
            return

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        edge_con_list_con = ""
        if edge_con_list is not None and len(edge_con_list) > 0:
            edge_con_list_con = " where " + " and ".join(edge_con_list)

        target_field_list_con = edge_info["src"] + "," + edge_info["dst"]
        if target_field_list is None or len(target_field_list) == 0:
            target_field_list_con += "," + ",".join(edge_info["fields"])
        elif len(target_field_list) > 0:
            target_field_list_con += "," + ",".join(target_field_list)

        sql = "select distinct " + target_field_list_con + " from " + db_table + edge_con_list_con
        logger.info("sql: " + sql)
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql).values.tolist()

        return res

    # def time_static_subgraph(self, subgraph_name, edge_name, edge_con_list, time_dimention='Day', data_type='list'):
    #
    #     graph_name = self.graph_name
    #     self.use_graph(subgraph_name)
    #     if "subgraph" not in self.graph_cfg or self.graph_cfg["subgraph"] != graph_name:
    #         logger.warning("not subgraph or graph-subgraph unmatched")
    #         self.use_graph(graph_name)
    #         return
    #     self.use_graph(subgraph_name)
    #     edge_info = self.graph_cfg["edges"][edge_name]['db'] + "." + self.graph_cfg["edges"][edge_name]['table']
    #     edge_con_list_con = ""
    #     if edge_con_list is not None and len(edge_con_list) > 0:
    #         edge_con_list_con = " where " + " and ".join(edge_con_list)
    #     sub_sql = "select distinct * from " + edge_info + edge_con_list_con
    #     sql = "select A.record_date, count() from " + "(" + sub_sql + ")" + " as A GROUP BY A.record_date"
    #     if time_dimention == 'Minute':
    #         sql = "select toStartOfMinute(A.record_time) as Minute,count() from " + "(" + sub_sql + ")" + " as A GROUP BY Minute"
    #     elif time_dimention == 'Hour':
    #         sql = "select toStartOfHour(A.record_time) as Hour, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Hour"
    #     elif time_dimention == 'Day':
    #         sql = "select formatDateTime(A.record_time,'%F') as Day, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Day"
    #     elif time_dimention == 'Month':
    #         sql = "select toStartOfMonth(A.record_time) as Month, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Month"
    #     elif time_dimention == 'Year':
    #         sql = "select formatDateTime(A.record_time,'%Y') as Year, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Year"
    #     # sql1 = "select formatDateTime(record_time,'%Y') as Year,toStartOfMonth(record_time),toStartOfDay(record_time),toStartOfHour(record_time),toStartOfMinute(record_time),record_time from  sub_anti_money_laundry2.accounts limit 10"
    #
    #     if data_type == "list":
    #         res = self.client.execute(sql)
    #     elif data_type == "df":
    #         res = self.client.query_dataframe(sql)
    #
    #     return res

    def count_edge_by_time(self, edge_name, edge_con_list, time_field, time_dimention='Day', data_type='df'):

        edge_info = self.graph_cfg["edges"][edge_name]
        db_table = edge_info["db"] + "." + edge_info["table"]

        edge_con_list_con = ""
        if edge_con_list is not None and len(edge_con_list) > 0:
            edge_con_list_con = " where " + " and ".join(edge_con_list)

        sub_sql = "select distinct * from " + db_table + edge_con_list_con
        sql = "select A." + time_field + ", count() from " + "(" + sub_sql + ")" + " as A GROUP BY A." + time_field
        if time_dimention == 'Minute':
            sql = "select toStartOfMinute(A." + time_field + ") as Minute, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Minute"
        elif time_dimention == 'Hour':
            sql = "select toStartOfHour(A." + time_field + ") as Hour, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Hour"
        elif time_dimention == 'Day':
            sql = "select formatDateTime(A." + time_field + ",'%F') as Day, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Day"
        elif time_dimention == 'Month':
            sql = "select toStartOfMonth(A." + time_field + ") as Month, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Month"
        elif time_dimention == 'Year':
            sql = "select formatDateTime(A." + time_field + ",'%Y') as Year, count() from " + "(" + sub_sql + ")" + " as A GROUP BY Year"
        # sql1 = "select formatDateTime(record_time,'%Y') as Year,toStartOfMonth(record_time),toStartOfDay(record_time),toStartOfHour(record_time),toStartOfMinute(record_time),record_time from  sub_anti_money_laundry2.accounts limit 10"

        print(sql)
        if data_type == "list":
            res = self.client.execute(sql)
        elif data_type == "df":
            res = self.client.query_dataframe(sql).values.tolist()

        return res











